from .models import TenantConfig, OidcBffConfig, OidcSharedSettings

def resolve_oidc_config(
    tenant_cfg: TenantConfig,
    shared: OidcSharedSettings,
) -> OidcBffConfig:
    
    if not tenant_cfg.oidc:
        raise ValueError(
            f"bff_oidc mancante per tenant '{tenant_cfg.tenant}'"
        )

    realm = tenant_cfg.oidc.realm
    issuer = f"{shared.base_url.rstrip('/')}/{realm}"

    tenant_public = tenant_cfg.oidc.public_base_url
    public = (tenant_public or shared.public_base_url).rstrip("/")
    
    return OidcBffConfig(
        realm=realm,
        issuer=issuer,
        authorization_endpoint=f"{issuer}/protocol/openid-connect/auth",
        token_endpoint=f"{issuer}/protocol/openid-connect/token",
        revocation_endpoint=f"{issuer}/protocol/openid-connect/revoke",
        end_session_endpoint=f"{issuer}/protocol/openid-connect/logout",
        jwks_uri=f"{issuer}/protocol/openid-connect/certs",
        client_id=shared.client_id,
        redirect_uri="http://localhost:8055/callback",
        post_logout_redirect_uri=f"{public}/",
        scope=shared.scope,
        token_endpoint_auth_method=shared.auth_method,
    )


from contextlib import asynccontextmanager
from contextvars import ContextVar

from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.pool import NullPool

_current_db_session: ContextVar[AsyncSession | None] = ContextVar(
    "current_db_session",
    default=None,
)

class TenantDatabaseManager:
    def __init__(self, config):
        self.config = config
        self._engines = {}
        self._sessionmakers = {}

    def _build_connect_args(self, tenant: str) -> dict:
        return {
            "options": f"-c application_name={self.config.app_name_prefix}:{tenant}"
        }

    def _engine_args(self):
        if self.config.mode == "testing":
            return {
                "echo": self.config.echo_sql,
                "poolclass": NullPool,
            }

        return {
            "echo": self.config.echo_sql,
            "pool_pre_ping": self.config.pool_pre_ping,
            "pool_size": self.config.pool_size,
            "max_overflow": self.config.max_overflow,
            "pool_timeout": self.config.pool_timeout,
        }

    def get_engine(self, tenant: str):
        if tenant not in self._engines:
            db_url = self.config.service.get(tenant).db.url.get_secret_value()

            engine = create_async_engine(
                db_url,
                connect_args=self._build_connect_args(tenant),
                **self._engine_args(),
            )

            self._engines[tenant] = engine

        return self._engines[tenant]

    def get_sessionmaker(self, tenant: str):
        if tenant not in self._sessionmakers:
            self._sessionmakers[tenant] = async_sessionmaker(
                self.get_engine(tenant),
                class_=AsyncSession,
                expire_on_commit=False,
            )
        return self._sessionmakers[tenant]

    @asynccontextmanager
    async def db_session_context(self, tenant: str):
        sessionmaker_factory = self.get_sessionmaker(tenant)

        async with sessionmaker_factory() as db_session:
            token = _current_db_session.set(db_session)
            try:
                yield db_session
            finally:
                _current_db_session.reset(token)

    def get_current_db_session(self) -> AsyncSession | None:
        return _current_db_session.get()

    async def dispose_all(self):
        for engine in self._engines.values():
            await engine.dispose()


from fastapi import Request
from fastapi.responses import JSONResponse
from sqlalchemy.exc import DBAPIError, OperationalError, StatementError


def make_tenant_session_dep(cfg):
    async def _dep(request: Request):
        async for _ in cfg.get_db_session(request):
            yield cfg.current_session()
    return _dep


async def database_error_middleware(request: Request, call_next):
    try:
        return await call_next(request)

    except OperationalError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Database non disponibile"},
        )

    except DBAPIError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Errore database"},
        )

    except StatementError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Errore query"},
        )


from contextlib import contextmanager
from contextvars import ContextVar, Token

from fastapi import HTTPException, Request

from .db_session import TenantDatabaseManager
from .service import TenantConfigService, TenantRuntimeConfigService

####logging di tenant e requestid
_current_request_id: ContextVar[str | None] = ContextVar(
    "current_request_id",
    default=None,
)

_current_tenant: ContextVar[str | None] = ContextVar(
    "current_tenant",
    default=None,
)

_current_user_id: ContextVar[str | None] = ContextVar(
    "current_user_id",
    default=None,
)

class MultitenantConfig:
    def __init__(
        self,
        tenants_source: str,
        *,
        #settings,
        tenant_header: str = "X-Tenant",
        user_id_header: str = "X-UserId",
        default_tenant: str | None = None,
        default_user_id: str | None = None,
        allowed_tenants: set[str] | None = None,
        strict_whitelist: bool = False,
        mode: str = "testing",
        echo_sql: bool = False,
        pool_size: int = 3,
        max_overflow: int = 0,
        pool_timeout: int = 5,
        pool_pre_ping: bool = True,
        app_name_prefix: str = "fastapi",
    ):
        #self.settings = settings

        self.tenant_header = tenant_header
        self.user_id_header = user_id_header
        self.default_tenant = default_tenant
        self.default_user_id = default_user_id

        self.allowed_tenants = allowed_tenants
        self.strict_whitelist = strict_whitelist

        self.mode = mode
        self.echo_sql = echo_sql
        self.pool_size = pool_size
        self.max_overflow = max_overflow
        self.pool_timeout = pool_timeout
        self.pool_pre_ping = pool_pre_ping
        self.app_name_prefix = app_name_prefix

        self.service = TenantConfigService(tenants_source)
        ###per ora commentata da vedere (https://chatgpt.com/c/69be4f40-1070-832a-ad71-6889e885bef3)
        #self.runtime = TenantRuntimeConfigService(self.service, settings)
        self.db = TenantDatabaseManager(self)
        
        
    # -------------------------
    # TENANT
    # -------------------------

    def get_tenant(self, request: Request) -> str:
        tenant = request.headers.get(self.tenant_header)
        
        if tenant and tenant.strip():
            tenant = tenant.strip().lower()
        else:
            tenant = self.default_tenant

        if not tenant:
            raise HTTPException(400, f"Header {self.tenant_header} mancante")

        if self.allowed_tenants and self.strict_whitelist:
            if tenant not in self.allowed_tenants:
                raise HTTPException(403, "Tenant non consentito")

        try:
            self.service.get(tenant)
        except Exception:
            raise HTTPException(404, f"Tenant '{tenant}' non trovato")

        return tenant
    
    # -------------------------
    # CONFIG
    # -------------------------

    def get_tenant_config(self, request: Request):
        tenant = self.get_tenant(request)
        return self.service.get(tenant)

    # -------------------------
    # RQUESTID (ContextVar)
    # -------------------------

    @contextmanager
    def request_id_context(self, request_id: str | None):
        token: Token = _current_request_id.set(request_id)
        try:
            yield
        finally:
            _current_request_id.reset(token)

    def current_request_id(self) -> str | None:
        return _current_request_id.get()

    # -------------------------
    # TENANT (ContextVar)
    # -------------------------

    @contextmanager
    def tenant_context(self, tenant: str | None):
        token: Token = _current_tenant.set(tenant)
        try:
            yield
        finally:
            _current_tenant.reset(token)

    def current_tenant(self) -> str | None:
        return _current_tenant.get()


    # -------------------------
    # USER ID (ContextVar)
    # -------------------------

    def get_user_id(self, request: Request) -> str | None:
        user_id = request.headers.get(self.user_id_header)
        return user_id.strip() if user_id and user_id.strip() else self.default_user_id

    @contextmanager
    def user_id_context(self, user_id: str | None):
        token: Token = _current_user_id.set(user_id)
        try:
            yield
        finally:
            _current_user_id.reset(token)

    def current_user_id(self) -> str | None:
        return _current_user_id.get()
    

    # -------------------------
    # DB SESSION (dependency)
    # -------------------------

    async def get_db_session(self, request: Request):
        tenant = self.get_tenant(request)
        user_id = self.get_user_id(request)
        
        with self.tenant_context(tenant), self.user_id_context(user_id):
            async with self.db.db_session_context(tenant):
                yield


    # -------------------------
    # HELPERS
    # -------------------------

    def current_db_session(self):
        session = self.db.get_current_db_session()
        if session is None:
            raise RuntimeError("Nessuna sessione DB attiva")
        return session
    
    async def dispose(self):
        await self.db.dispose_all()


from __future__ import annotations

from pathlib import Path

from pydantic import SecretStr

from .models import (
    EmailConfig,
    OidcBffConfig,
    OidcSharedSettings,
    PagopaConfig,
    ProtocolloConfig,
    TenantConfig,
)


def _read_secret_file(path: str | None) -> str | None:
    if not path:
        return None

    value = Path(path).read_text(encoding="utf-8").strip()
    return value or None


def _coalesce(*values):
    for value in values:
        if value is not None:
            return value
    return None


def resolve_oidc_config(
    tenant_cfg: TenantConfig,
    shared: OidcSharedSettings,
) -> OidcBffConfig:
    tenant_oidc = tenant_cfg.bff_oidc

    realm = tenant_oidc.realm
    client_id = _coalesce(tenant_oidc.client_id, shared.client_id)
    public_base_url = _coalesce(
        tenant_oidc.public_base_url,
        shared.public_base_url,
    )
    redirect_path = _coalesce(
        tenant_oidc.redirect_path,
        shared.redirect_path,
    )
    scope = _coalesce(
        tenant_oidc.scope,
        shared.scope,
    )
    auth_method = _coalesce(
        tenant_oidc.auth_method,
        shared.auth_method,
    )

    base_url = shared.base_url.rstrip("/")
    issuer = f"{base_url}/realms/{realm}"

    redirect_uri = f"{public_base_url.rstrip('/')}{redirect_path}"
    post_logout_redirect_uri = public_base_url.rstrip("/")

    return OidcBffConfig(
        realm=realm,
        issuer=issuer,
        authorization_endpoint=f"{issuer}/protocol/openid-connect/auth",
        token_endpoint=f"{issuer}/protocol/openid-connect/token",
        revocation_endpoint=f"{issuer}/protocol/openid-connect/revoke",
        end_session_endpoint=f"{issuer}/protocol/openid-connect/logout",
        jwks_uri=f"{issuer}/protocol/openid-connect/certs",
        client_id=client_id,
        redirect_uri=redirect_uri,
        post_logout_redirect_uri=post_logout_redirect_uri,
        scope=scope,
        token_endpoint_auth_method=auth_method,
    )


def resolve_email_config(
    tenant_cfg: TenantConfig,
    settings,
) -> EmailConfig:
    tenant_email = tenant_cfg.email

    default_password: str | None = None
    if getattr(settings, "smtp_password", None):
        default_password = settings.smtp_password
    elif getattr(settings, "smtp_password_file", None):
        default_password = _read_secret_file(settings.smtp_password_file)

    tenant_password = (
        tenant_email.password.get_secret_value()
        if tenant_email and tenant_email.password
        else None
    )

    password = _coalesce(tenant_password, default_password)
    if password is None:
        raise ValueError(
            f"Configurazione email incompleta per tenant '{tenant_cfg.tenant}': "
            "password non disponibile"
        )

    return EmailConfig(
        smtp_host=_coalesce(
            tenant_email.smtp_host if tenant_email else None,
            settings.smtp_host,
        ),
        smtp_port=_coalesce(
            tenant_email.smtp_port if tenant_email else None,
            settings.smtp_port,
        ),
        username=_coalesce(
            tenant_email.username if tenant_email else None,
            settings.smtp_username,
        ),
        password=SecretStr(password),
        use_tls=_coalesce(
            tenant_email.use_tls if tenant_email else None,
            settings.smtp_use_tls,
        ),
        from_address=_coalesce(
            tenant_email.from_address if tenant_email else None,
            getattr(settings, "smtp_from_address", None),
        ),
    )


def resolve_pagopa_config(
    tenant_cfg: TenantConfig,
    settings,
) -> PagopaConfig:
    tenant_pagopa = tenant_cfg.pagopa

    default_password = _coalesce(
        getattr(settings, "pagopa_password", None),
        _read_secret_file(getattr(settings, "pagopa_password_file", None)),
    )
    default_print_password = _coalesce(
        getattr(settings, "pagopa_print_password", None),
        _read_secret_file(getattr(settings, "pagopa_print_password_file", None)),
    )

    tenant_password = (
        tenant_pagopa.password.get_secret_value()
        if tenant_pagopa and tenant_pagopa.password
        else None
    )
    tenant_print_password = (
        tenant_pagopa.print_password.get_secret_value()
        if tenant_pagopa and tenant_pagopa.print_password
        else None
    )

    password = _coalesce(tenant_password, default_password)
    print_password = _coalesce(tenant_print_password, default_print_password)

    if password is None:
        raise ValueError(
            f"Configurazione PagoPA incompleta per tenant '{tenant_cfg.tenant}': "
            "password non disponibile"
        )

    if print_password is None:
        raise ValueError(
            f"Configurazione PagoPA incompleta per tenant '{tenant_cfg.tenant}': "
            "print_password non disponibile"
        )

    return PagopaConfig(
        username=_coalesce(
            tenant_pagopa.username if tenant_pagopa else None,
            getattr(settings, "pagopa_username", None),
        ),
        password=SecretStr(password),
        print_username=_coalesce(
            tenant_pagopa.print_username if tenant_pagopa else None,
            getattr(settings, "pagopa_print_username", None),
        ),
        print_password=SecretStr(print_password),
        service_url=_coalesce(
            tenant_pagopa.service_url if tenant_pagopa else None,
            getattr(settings, "pagopa_service_url", None),
        ),
        codipa=_coalesce(
            tenant_pagopa.codipa if tenant_pagopa else None,
            getattr(settings, "pagopa_codipa", None),
        ),
    )


def resolve_protocollo_config(
    tenant_cfg: TenantConfig,
    settings,
) -> ProtocolloConfig:
    tenant_protocollo = tenant_cfg.protocollo

    default_password = _coalesce(
        getattr(settings, "protocollo_password", None),
        _read_secret_file(getattr(settings, "protocollo_password_file", None)),
    )

    tenant_password = (
        tenant_protocollo.password.get_secret_value()
        if tenant_protocollo and tenant_protocollo.password
        else None
    )

    password = _coalesce(tenant_password, default_password)
    if password is None:
        raise ValueError(
            f"Configurazione protocollo incompleta per tenant '{tenant_cfg.tenant}': "
            "password non disponibile"
        )

    return ProtocolloConfig(
        base_url=_coalesce(
            tenant_protocollo.base_url if tenant_protocollo else None,
            getattr(settings, "protocollo_base_url", None),
        ),
        username=_coalesce(
            tenant_protocollo.username if tenant_protocollo else None,
            getattr(settings, "protocollo_username", None),
        ),
        password=SecretStr(password),
    )


from typing import Any, Generic, TypeVar

from pydantic import BaseModel
from sqlalchemy import select, func, delete as sa_delete
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect

ModelType = TypeVar("ModelType")
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class MissingSessionError(RuntimeError):
    pass


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: type[ModelType], *, config=None):
        self.model = model
        self.config = config

    # -------------------------
    # SESSION
    # -------------------------

    def _db(self, db: AsyncSession | None) -> AsyncSession:
        if db:
            return db

        if self.config and hasattr(self.config, "current_db_session"):
            return self.config.current_db_session()

        raise MissingSessionError(f"No DB session for {self.model.__name__}")


    # -------------------------
    # USER (audit)
    # -------------------------

    def _current_user_id(self) -> str | None:
        if self.config and hasattr(self.config, "current_user_id"):
            return self.config.current_user_id()
        return None

    def _apply_create_audit(self, obj):
        user_id = self._current_user_id()
        if not user_id:
            return

        if hasattr(obj, "created_by") and not getattr(obj, "created_by", None):
            obj.created_by = user_id

        if hasattr(obj, "updated_by"):
            obj.updated_by = user_id

    def _apply_update_audit(self, obj):
        user_id = self._current_user_id()
        if user_id and hasattr(obj, "updated_by"):
            obj.updated_by = user_id

    # -------------------------
    # PK
    # -------------------------

    @property
    def pk_column(self):
        mapper = inspect(self.model)
        pk = mapper.primary_key

        if len(pk) != 1:
            raise ValueError("Only single PK supported")

        return pk[0]

    # -------------------------
    # GET
    # -------------------------

    async def get(self, id: Any, *, db: AsyncSession | None = None):
        db = self._db(db)
        return await db.get(self.model, id)

    async def get_multi(
        self,
        *,
        offset: int = 0,
        limit: int = 100,
        db: AsyncSession | None = None,
    ):
        db = self._db(db)

        stmt = select(self.model).offset(offset).limit(limit)
        result = await db.execute(stmt)
        return result.scalars().all()

    async def count(self, *, db: AsyncSession | None = None):
        db = self._db(db)

        stmt = select(func.count()).select_from(self.model)
        result = await db.execute(stmt)
        return result.scalar_one()

    # -------------------------
    # CREATE
    # -------------------------

    async def create(
        self,
        *,
        obj_in: CreateSchemaType | dict,
        db: AsyncSession | None = None,
        commit: bool = True,
    ):
        db = self._db(db)

        data = (
            obj_in.model_dump(exclude_unset=True)
            if isinstance(obj_in, BaseModel)
            else obj_in
        )

        db_obj = self.model(**data)

        self._apply_create_audit(db_obj)

        db.add(db_obj)

        if commit:
            await db.commit()
            await db.refresh(db_obj)
        else:
            await db.flush()

        return db_obj

    # -------------------------
    # UPDATE
    # -------------------------

    async def update(
        self,
        *,
        db_obj,
        obj_in,
        db: AsyncSession | None = None,
        commit: bool = True,
    ):
        db = self._db(db)

        data = (
            obj_in.model_dump(exclude_unset=True)
            if isinstance(obj_in, BaseModel)
            else obj_in
        )

        for k, v in data.items():
            setattr(db_obj, k, v)

        self._apply_update_audit(db_obj)

        db.add(db_obj)

        if commit:
            await db.commit()
            await db.refresh(db_obj)
        else:
            await db.flush()

        return db_obj

    # -------------------------
    # DELETE
    # -------------------------

    async def delete(
        self,
        *,
        id: Any,
        db: AsyncSession | None = None,
        commit: bool = True,
    ):
        db = self._db(db)

        obj = await db.get(self.model, id)
        if not obj:
            return None

        await db.delete(obj)

        if commit:
            await db.commit()
        else:
            await db.flush()

        return obj

    async def delete_where(
        self,
        *,
        where,
        db: AsyncSession | None = None,
        commit: bool = True,
    ):
        db = self._db(db)

        stmt = sa_delete(self.model).where(where)
        result = await db.execute(stmt)

        if commit:
            await db.commit()
        else:
            await db.flush()

        return result.rowcount or 0